;; Transport Layer
;; As for now we run all on one computer, the network connection is simulated using channels (PLT Ref 10.2.2)
;; hence the name dummy-transport-layer
;; It can however be easily modified to use the network: replace the channels we use by sockets. 

(require scheme/serialize)

(define-serializable-struct remote-args (fstub fargs asps jps caller))

;shadowed in remote-exec 
(define target (make-parameter #f))
(define (get-target-host) 
  (target))

;; Execute call on a remote host
(define (remote-exec stub args A jp)
  (let* ((target-host (fun-stub-host stub))
         (name  (host-name target-host))
         (port  (host-port target-host))
         (queue (get-comms-queue name port)))
    ;; set target host for the extent of serialization
    (parameterize ((target target-host))
      (let* ((new-args (map (lambda (a) 
                              (if (closure? a) (copy-function a A jp) a)) ;; FIX: lists
                            args))
             ;; as before
             (single-arg (make-remote-args stub new-args A jp (get-current-host))))
        ;; pass the serialized call
        (channel-put queue (serialize single-arg))))
    ;; get result back
    (deserialize (channel-get (channel-get queue))))) 


;; Accept one incoming call, call is executed in a separate thread
(define (accept-call queue)
  (server-printf "Server waiting for message~n")
  (let ((call (channel-get queue))
        (call-specific-queue (make-channel))
        (parent-thread (current-thread)))
    (channel-put queue call-specific-queue)
    (thread
     (lambda ()
       (begin (register-handler-thread (current-thread) parent-thread)
              (channel-put call-specific-queue (receive-incoming-call call))
              (unregister-handler-thread (current-thread)))))))

(define (receive-incoming-call serialized-single-arg)
  (server-printf "Executing incoming function call ")
  (let* ((single-arg (deserialize serialized-single-arg))
         (stub   (remote-args-fstub single-arg)) 
         (args   (remote-args-fargs single-arg))
         (A      (remote-args-asps single-arg)) 
         (jp     (remote-args-jps single-arg))
         (client (remote-args-caller single-arg))
         (f      (lookup-exported-function (fun-stub-name stub))) 
         (fake   (server-printf "to function ~s~n" (fun-stub-name stub)))
         (result (exec-fun f args A jp)))
    (parameterize ((target client))
      (serialize (if (closure? result)  ;; FIX: lists
                     (copy-function result A jp) 
                     result)))))

(define comms-queue-store '())

(define (make-comms-queue host port)
  (let ((channel (make-channel)))
    (set! comms-queue-store (cons (cons (string-append host ":" port) channel) comms-queue-store))
    channel))

(define (get-comms-queue host port)
  (let ((binding (assoc (string-append host ":" port) comms-queue-store)))
    (if (eq? binding #f)
        (error 'get-comms-queue "Looking up queue to host ~s" (string-append host ":" port))
        (cdr binding))))               


