(load "ds-fp.scm")
(load "test.scm")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;; TESTS for the non-distributed interpreter ;;;;;;;;;;;;

;; These are based on the tests for the AOSD 2008 interpreter (non-distributed deployment strategies)
;; The 2008 tests have been modified to use the testing framework
;; Tests were added for cflow, cflowbelow, and deploy-on

;;expect: 10
(test (run '10) 10)
;;expect: 15
(test (interp (parse 'x) (new-env (list 'x) (list 15)) '() '()) 15)
;;expect: 16
(test (interp (parse '(+ 1 x)) (new-env (list 'x) (list 15)) '() '()) 16)
;;expect: #<struct:closure>
(test/pred (run '(fun (x y) (+ x y))) closure?)
;;expect: 25
(test (interp (parse '(let ((y 10)) (+ y x))) (new-env (list 'x) (list 15)) '() '()) 25)
;;expect: 2
(test (run '(let ((f (fun (x y) (+ x y))))
          (write (f 3 4))
          (f 1 1))) 2)
;;expect:3
(test (run '(begin
              (define x 3) 
              x)) 3)
;;expect:4
(test (run '(begin
              (define x 3) 
              (set! x 4)
              x)) 4)
;; testing list/assq
;; expect: c
(test (run '(let ((map '((1 . a) (2 . b) (3 . c) (4 . d))))
              (cdr (assq 3 map)))) 'c)

;; Expressing AspectScheme semantics with deployment strategies
(printf "~n---dynamically-scoped: aspect applies~n")
(run '(let ((open-file (fun (file) (printf "->opening ~a~n" file))))
        (let ((apply-to-brussels (fun (f) (f "brussels")))) 
          (deploy as-dynamic 
                  (list (call open-file) (fun (jp) (printf "trace~n")))
                  (apply-to-brussels open-file)))))

(printf "~n---statically-scoped: aspect does not apply~n")
(run '(let ((open-file (fun (file) (printf "->opening ~a~n" file))))
        (let ((apply-to-brussels (fun (f) (f "brussels"))))
          (deploy as-static
                  (list (call open-file) (fun (jp) (printf "trace~n")))
                  (apply-to-brussels open-file)))))

(printf "~n---testing cflow: apply once~n")
(run '(let ((f (fun () 'done)))
        (let ((g (fun () (f))))
          (deploy as-dynamic
                  (list (&& (exec f) (! (cflow (call g)))) 
                        (fun (jp) (printf "trace~n")))
                  (g)
                  (f)
                  ))))


(printf "~n---testing cflowbelow: apply once~n")
(run '(begin (define fact (fun (n) 
                               (if (eq? 0 n) 1
                                   (* n (fact (- n 1))))))
             (deploy as-dynamic
                     (list (&& (exec fact) (! (cflowbelow (exec fact))))
                           (fun (jp) (printf "trace~n")))
                     (fact 4))))

(printf "~n---testing cflowbelow: apply 4 times~n")
(run '(begin (define fact (fun (n) 
                               (if (eq? 0 n) 1
                                   (* n (fact (- n 1))))))
             (deploy as-dynamic
                     (list (&& (exec fact) (cflowbelow (exec fact)))
                           (fun (jp) (printf "trace~n")))
                     (fact 4))))

(printf "~n---statically-scoped: aspect applies~n")
(run '(let ((open-file (fun (file) (printf "->opening ~a~n" file))))
        (let ((traced-open  (deploy as-static
                                    (list (call open-file) (fun (jp) (printf "trace~n")))
                                    (fun (f) (open-file f)))))
          (traced-open "brussels"))))


(printf "~n---statically-scoped (using deploy-on): aspect applies~n")
(run '(let ((open-file (fun (file) (printf "->opening ~a~n" file))))
        (let ((traced-open (fun (f) (open-file f)))) ;; function created beforehand
          (deploy-on (list f f t) ;; does not matter
                     (list (call open-file) (fun (jp) (printf "trace~n")))
                     traced-open) ;; deploy aspect in function aspect environment
          (traced-open "brussels"))))

(printf "~n---dynamically-scoped: aspect does not apply~n")
(run '(let ((open-file (fun (file) (printf "->opening ~a~n" file))))
        (let ((traced-open (deploy as-dynamic 
                                   (list (call open-file) (fun (jp) (printf "trace~n")))
                                   (fun (f) (open-file f)))))
          (traced-open "brussels"))))

;;; parameterized propagation functions
(printf "~n---propagate on stack only if applying apply-to-paris: once~n")
(run 
 '(let ((open-file (fun (file) (printf "->opening ~a~n" file))))
    (let ((apply-to-brussels (fun (f) (f "brussels")))
          (apply-to-paris    (fun (f) (f "paris")))) 
      (deploy (list (fun (jp) (eq? (jp-fun jp) apply-to-paris)) t t) 
              (list (fun (jp) (eq? (jp-fun jp) open-file)) (fun (jp) (printf "trace~n")))
              (apply-to-brussels open-file)
              (apply-to-paris open-file)
              ))))

(printf "~n---propagate on stack depending on relative stack depth~n")
(run
 '(let ((stack-depth-<= (fun (n jp) 
                             (fun (jp2) 
                                  (<= (- (jp-depth jp2) (jp-depth jp)) n))))
        (open-file (fun (file) (printf "->opening ~a~n" file))))
    (let ((open-log (fun () (open-file "log"))))
      (let ((open-both (fun (file) 
                            (open-file file)
                            (open-log))))
        (deploy (list (stack-depth-<= 2 (this-jp)) f t) 
                (list (call open-file) (fun (jp) (printf "trace~n")))
                (open-both "brussels"))))))


;;; join point filters
(printf "~n---join point filter~n")
(run 
 '(let ((open-file (fun (file) (printf "->opening ~a~n" file)))
        (aspect (list (fun (jp) (eq? (jp-kind jp) 'call)) 
                      (fun (jp) (printf "trace~n"))))) ;; trace all calls
    
    (let ((apply-to-brussels (fun (f) (f "brussels")))
          (apply-to-paris    (fun (f) (f "paris")))) 
      
      (printf "here match twice~n")
      (deploy (list t f t) 
              aspect
              (apply-to-paris open-file))
      
      (printf "here match once~n")
      (deploy (list t f (fun (jp) (eq? (jp-fun jp) apply-to-paris))) 
              aspect 
              (apply-to-paris open-file)))))